#                                                                                         
#    Connect timeout with SELECT - bornSoft                                               
#    --------------------------------------                                               
#                                                                                         
#    You need to run this script as "VDS6 CONSOLE EXECUTABLE" in the Project Manager.     
#                                                                                         
#    This is a test-script to evaluate the Winsock SELECT function and the                
#    corresponding FD_SET macros.                                                         
#                                                                                         
#    A connection will be made to the POP3 server "pop.gmx.net" and its capabilities      
#    will be requested. After that the connection will be closed.                         
#                                                                                         
#    To run this script successfully you need to select "VDS6 Console Executable" on      
#    the Project tab in the Project Manager of the IDE. The files bsWinSock.dsu and       
#    bsWinSock.h must reside in the same directory as the script and of course you        
#    need an established internet connection.                                             
#                                                                                         
#    To intendedly produce a connection timeout just change the %%Port variable to        
#    something else than 110.                                                             
#                                                                                         
#    The SELECT timeout is set to 1 second an this is repeated 10 times, so there is      
#    total timeout for the connection request of 10 seconds.                              
#                                                                                         
#    To close the console window at the end hit ENTEER.                                   
#                                                                                         

  #INCLUDE bsWinSock.dsu

  #DEFINE Function,select
  #DEFINE Function,Create_FD_SET
  #DEFINE Function,FD_ZERO
  #DEFINE Function,FD_SET
  #DEFINE Function,FD_ISSET
  #DEFINE Function,FD_CLR

  option decimalsep,.

  #    Server to connect  
  %%Server = pop.mail.yahoo.de
  %%Port = 110

  #    Initialize     
  bsWinSock INITIALIZE

  #    Startup Windows Sockets 
  %R = @bsWinSock(WSAStartup)

  writeline console," "
  writeline console," Test - Connect timeout with SELECT"
  writeline console," ----------------------------------"@cr()@lf()

  #    Create TCP socket           
  %%Socket = @bsWinSock(socket,TCP)

  #    More test-sockets to fill FD_SET 
  %%Sock_1 = @bsWinSock(socket,TCP)
  %%Sock_2 = @bsWinSock(socket,TCP)
  %%Sock_3 = @bsWinSock(socket,TCP)

  #    Make the socket nonblocking               
  %R = @bsWinSock(ioctlsocket,%%Socket,FIONBIO,1)
  writeline console," Socket created ["%%Socket"]."@cr()@lf()
  writeline console," Setting socket to non-blocking mode."@cr()@lf()

  #    Resolve hostname of the server to IPv4  
  %%Server = @bsWinSock(gethostbyname,%%Server)

  #    Connect to the server                       
  %R = @bsWinSock(connect,%%Socket,%%Server,%%Port)
  write console," Connecting to "%%Server" on port "%%Port"  "

  #    Errorcheck                    
  %%CONNECT_ERROR = @bsWinSock(ERROR)
  #    If error = WSAEWOULDBLOCK (10035)      
  if @equal(%%CONNECT_ERROR,%%_WSAEWOULDBLOCK)

      %T = 0
      repeat

          #    Create FD_SET for writeability check 
          %%WRITE_FD_SET = @FD_ZERO(%%WRITE_FD_SET)
 
          #    Add the Socket to the FD_SET                  
          %%WRITE_FD_SET = @FD_SET(0,%%Socket,%%WRITE_FD_SET)
 
          #    Add test-sockets to the FD_SET                
          %%WRITE_FD_SET = @FD_SET(0,%%Sock_1,%%WRITE_FD_SET)
          %%WRITE_FD_SET = @FD_SET(0,%%Sock_2,%%WRITE_FD_SET)
          %%WRITE_FD_SET = @FD_SET(0,%%Sock_3,%%WRITE_FD_SET)

        #  %%WRITE_FD_SET = @binary(DWORD,1)@binary(DWORD,%%Socket)

          #    select timeout 1 second                 
          %%connected = @select(0,0,%%WRITE_FD_SET,0,1)
          %%SELECT_ERROR = @LIB(ws2_32,WSAGetLastError,INT:)

          #    Check if socket is within FD_SET    
          %%S1 = @FD_ISSET(0,%%Sock_1,%%WRITE_FD_SET)
          %%S2 = @FD_ISSET(0,%%Sock_2,%%WRITE_FD_SET)
          %%S3 = @FD_ISSET(0,%%Sock_3,%%WRITE_FD_SET)
          %S = @FD_ISSET(0,%%Socket,%%WRITE_FD_SET)
 
          #                                                                   
          #    This all works pretty good, except the FD_SET things.          
          #    If I add the socket to %%WRITE_FD_SET the FD_ISSET function    
          #    which in fact is the WinSock function "__WSAFDIsSet" always    
          #    returns true, even if the connection failed.                   
          #                                                                   
          #    The Problem must be in the data structure of FD_SET, so that   
          #    SELECT can not update it properly. But I'm wondering why       
          #    the return value of SELECT is correct and no socket errors     
          #    are reported, what should be indicating that the structure is  
          #    correct. Very strange ...                                      
          #                                                                   
          #    If all would be ok, the following UNTIL ... should be:         
          #                                                                   
          #    until @greater(@FD_ISSET(0,%%Socket,%%WRITE_FD_SET),0)         
          #                                                                   

          wait 1
          write console,"# "
          %T = @succ(%T)
      until @greater(%%connected,0) @greater(%T,10)

      #    Exit on connect timeout         
      if @greater(%T,10) @zero(%%Connected)
          writeline console,@cr()@lf()@cr()@lf()" Connect timed out (10 sec)."@cr()@lf()
          goto Close
      end

    else
      if @greater(%%CONNECT_ERROR,0)
          writeline console," A connection error occured"
      end
  end

  #    Make the socket blocking again                                                       
  writeline console,@cr()@lf()@cr()@lf()" Connected to "%%Server" on port "%%Port.@cr()@lf()
  writeline console," Setting socket back to blocking mode."@cr()@lf()@cr()@lf()
  %R = @bsWinSock(ioctlsocket,%%Socket,FIONBIO,0)

  #    Read welcome message from the server 
  %R = @bsWinSock(recv,%%Socket,512)
  writeline console," Server: "@bsWinSock(ReadBuffer)

  #    Send CAPA command to the server           
  %R = @bsWinSock(send,%%Socket,"CAPA"@cr()@lf())
  writeline console," Client: CAPA"@cr()@lf()

  #    Read response to CAPA command 
  %R = @bsWinSock(recv,%%Socket,512)
  writeline console," Server: "@bsWinSock(ReadBuffer)

:Quit
  #    Send QUIT command to the server           
  %R = @bsWinSock(send,%%Socket,"QUIT"@cr()@lf())
  writeline console," Client: QUIT"@cr()@lf()

  #    Read response to QUIT command 
  %R = @bsWinSock(recv,%%Socket,512)
  writeline console," Server: "@bsWinSock(ReadBuffer)

:Close
  #    Finish and cleanup stuff       
  %R = @bsWinSock(shutdown,%%Socket,2)
  writeline console," Connection closed."@cr()@lf()

  %R = @bsWinSock(shutdown,%%Sock_1,2)
  %R = @bsWinSock(shutdown,%%Sock_2,2)
  %R = @bsWinSock(shutdown,%%Sock_3,2)
 
  %R = @bsWinSock(closesocket,%%Socket)
 
  %R = @bsWinSock(closesocket,%%Sock_1)
  %R = @bsWinSock(closesocket,%%Sock_2)
  %R = @bsWinSock(closesocket,%%Sock_3)
 
  %R = @bsWinSock(WSACleanup)
 
  gosub DebugInfo

  writeline console," Program terminated."@cr()@lf()
  writeline console," Hit RETURN to close console window."@cr()@lf()@cr()@lf()
  while @not(@read(console,1))
      wait 0.1
  wend
  stop



#                                                                                         
#    Functions and Macros                                                                 
#                                                                                         

  #  typedef struct fd_set {           
  #  u_int  fd_count;                  
  #  SOCKET fd_array[FD_SETSIZE];      
  #  } fd_set;                         

:FD_ZERO
  #  Macro FD_ZERO ( by default FD_SETSIZE = 64 )      
  %R = @binary(DWORD,0)@fill(@prod(%%_FD_SETSIZE,4),,z)
  exit %R

:FD_ISSET  
  #  This works !! - but giving false positives! 

  #  Macro FD_ISSET (Returns 0 or greater than 0)   
  %R = @LIB(ws2_32,__WSAFDIsSet,INT:,%2,@addr("%3"))
  %%bsWinSock_SOCKETERROR = @LIB(ws2_32,WSAGetLastError,INT:)
  exit %R

:FD_SET
  #    Macro FD_SET()       
  #                         
  #    %S -> Socket / %F -> FD_SET / %C -> fd_count / %A -> fd_array 
  if @not(%2) @not(%3) @less(%2,1) @less(@len(%3),4)
      %%bsWinSock_SOCKETERROR = 10022
      Error -13
      exit -1
  end
  %S = %2
  %F = %3
  #    %C holds fd_count    
  %C = @val(@substr(%F,1,4))
  if @greater(%C,0)
      #    %A holds fd_array    
      %A = @strdel(%F,1,4)
      #    Sorted to remove doubles 
      %L = @new(list,sorted)
      while @greater(@len(%A),0)
          list add,%L,@val(@substr(%A,1,4))
          %A = @strdel(%A,1,4)
      wend
      #    Add the new socket 
      list add,%L,%S
      #    Update fd_Count 
      %C = @count(%L)
      list seek,%L,0
      while @item(%L)
          %A = %A@binary(DWORD,@item(%L))
          list delete,%L
      wend
      list close,%L
    else
      #    If it's the first one 
      %A = @binary(DWORD,%S)
      %C = @succ(%C)
  end
  %R = @binary(DWORD,%C)%A
  #   This doesn't work               
  #%R = %R@fill(@diff(260,@len(%R)),,z)
  exit %R


:select
  #    READ_FD_SET    
  if @not(@zero(%2))
      %A = @addr("%2")
    else
      %A = 0
  end
  #    WRITE_FD_SET   
  if @not(@zero(%3))
      %B = @addr("%3")
    else
      %B = 0
  end
  #    EXCEPT_FD_SET  
  if @not(@zero(%4))
      %C = @addr("%4")
    else
      %C = 0
  end
  #  struct timeval {                  
  #  long    tv_sec    // seconds      
  #  long    tv_usec   // microseconds 
  #  };                                
  if @equal(%5,NULL)
      %T = 0
    elsif @greater(%5,0)
      %T = @fmul(%5,1000000)
      %%tv_sec = @binary(DWORD,0)
      %%tv_usec = @binary(DWORD,%T)
      %T = %%tv_sec%%tv_usec
    else
      %T = @binary(DWORD,0)@binary(DWORD,0)
  end
  %D = @addr("%T")
  #    select , nfds (ignored) , *readfds , *writefds , *exceptfds , *timeout 
  %R = @LIB(ws2_32,select,INT:,NIL,%A,%B,%C,%D)
  %%bsWinSock_SOCKETERROR = @LIB(ws2_32,WSAGetLastError,INT:)
  exit %R

:DebugInfo
  writeline console," Debug-Info:"@cr()@lf()
  %R = %%CONNECT_ERROR
      gosub bsWinsock_WSAErrorList
  writeline console," %%CONNECT_ERROR  = "%R
  %R = %%SELECT_ERROR
      gosub bsWinsock_WSAErrorList
  writeline console," %%SELECT_ERROR   = "%R
  writeline console," %%connected      = "%%connected
  writeline console," "
  writeline console," Return values from FD_ISSET function:"@cr()@lf()
  writeline console," %S               = "%S
  writeline console," %%S1             = "%%S1
  writeline console," %%S2             = "%%S2
  writeline console," %%S3             = "%%S3
  writeline console," "
  exit
